from fastapi import FastAPI
from typing import Any, Dict, List, Union
from rockydb.rocky import RockyDB
from pydantic import BaseModel


class Settings(BaseModel):
    db_path: str = "../database/"


app_settings = Settings()
app = FastAPI()


@app.get("/")
async def status():
    return {"status": "online"}


@app.get("/settings/info")
async def settings_info():
    return {"db_path": app_settings.db_path}


@app.post("/settings/update")
async def update_settings(settings: Settings):
    app_settings.db_path = settings.db_path

    return {"db_path": app_settings.db_path}


@app.get("/{collection_name}/search/")
async def search(collection_name, type: str, field: str, value: str):
    collection = RockyDB(app_settings.db_path).collection(collection_name)
    documents = collection.search(field=field, value=value, type=type)

    return {"documents": documents}


@app.get("/{collection_name}/document/{doc_id}")
async def get_document(collection_name: str, doc_id):
    collection = RockyDB(app_settings.db_path).collection(collection_name)
    document = collection.get(str(doc_id))

    return {"document": document}


@app.delete("/{collection_name}/delete/{doc_id}")
async def delete_document(collection_name: str, doc_id):
    collection = RockyDB(app_settings.db_path).collection(collection_name)
    did_delete = collection.delete([doc_id])

    return {"didDelete": did_delete}


@app.post("/{collection_name}/insert/")
async def insert_doc(collection_name: str, document: Union[List, Dict, Any] = None):
    collection = RockyDB(app_settings.db_path).collection(collection_name)
    doc_id = collection.insert(document)

    return {"_id": doc_id}



# the encoding/decoding code for all data types will go here. Will need
# to a more faster json library in the future.

import json
import struct
from typing import Union


def encode_this(x: Union[int, float, bool, str, list]) -> bytes:
    if type(x) == int:
        xbytes = encode_int(x)
        return xbytes

    if type(x) == float:
        xbytes = encode_float(x)
        return xbytes

    if type(x) == bool:
        xbytes = encode_bool(x)
        return xbytes

    if type(x) == str:
        xbytes = encode_str(x)
        return xbytes

    if type(x) == list:
        xbytes = encode_array(x)
        return xbytes

    # raise some sort of error if its none of the listed types


def decode_this(
    data_type: Union[int, float, bool, str, list], xbytes: bytes
) -> Union[int, float, bool, str, list]:
    if data_type == int:
        x = decode_int(xbytes)
        return x

    if data_type == float:
        x = decode_float(xbytes)
        return x

    if data_type == bool:
        x = decode_bool(xbytes)
        return x

    if data_type == str:
        x = decode_str(xbytes)
        return x

    if data_type == list:
        x = decode_array(xbytes)
        return x


# int
def encode_int(x: int) -> bytes:
    byte_length = (x.bit_length() + 7) // 8
    return x.to_bytes(byte_length, "big")


def decode_int(xbytes: bytes) -> int:
    return int.from_bytes(xbytes, "big")


# float
def encode_float(x: float) -> bytes:
    return struct.pack("<d", x)


def decode_float(xbytes: bytes) -> float:
    return struct.unpack("<d", xbytes)[0]


# boolean
def encode_bool(x: bool) -> bytes:
    return struct.pack("?", x)


def decode_bool(xbytes: bytes) -> bool:
    return struct.unpack("?", xbytes)[0]


# str
def encode_str(x: str) -> bytes:
    return bytes(x, "utf-8")


def decode_str(xbytes: bytes) -> str:
    return str(xbytes.decode())


# array
def encode_array(arr: list) -> bytes:
    # make sure list is not nested
    if any(isinstance(i, list) for i in arr):
        return None

    arr_str = json.dumps(arr)
    arr_bytes = arr_str.encode("utf-8")

    return arr_bytes


def decode_array(arr: bytes) -> list:
    arr_decoded = arr.decode("utf-8")
    arr = json.loads(arr_decoded)

    return arr



from pathlib import Path
from rockydb.collection import Collection
import shutil


class RockyDB:
    def __init__(self, path: str = "database/"):
        self.path = path
        self._create_dir(self.path)

    def _create_dir(self, dir_path: str, with_meta: bool = False) -> bool:
        if Path(dir_path).is_dir():
            return False

        # make directory
        db_path = Path(dir_path)
        db_path.mkdir(parents=True, exist_ok=True)

        return True

    def collection(self, name: str):
        # add to meta.json file that belongs to db (outside of all collections)
        return Collection(self.path, name)

    def clean_up(self):
        database_path = Path(self.path)
        shutil.rmtree(self.path, ignore_errors=True)

        # database_files = list(database_path.iterdir())

        # for filename in database_files:
        #     filename.unlink()



# from typing import Optional
# import tantivy
# import json
# from pathlib import Path
# import rockydb.encoding as encoding
# from rocksdict import ReadOptions, Rdict


# class Index:
#     def __init__(
#         self,
#         db_path: str,
#         collection: Rdict,
#         collection_name: str,
#         name: str,
#         fields: Optional[list] = None,
#         encoding_types: dict = None,
#     ):

#         self.name = name
#         self.fields = fields
#         self.db_path = db_path
#         self.collection = collection
#         self.collection_name = collection_name
#         self.encoding_types = encoding_types

#     def _decode_value(self, value: bytes):
#         if not value:
#             return None

#         decoded_data_type = self.encoding_types[value[0]]
#         decoded_value = encoding.decode_this(decoded_data_type, value[1:])

#         return decoded_value

#     def _get(self, key: bytes):
#         value = self.collection[key]
#         return self._decode_value(value)

#     def get(self, id: str):
#         document = {}

#         key = encoding.encode_str(self.collection_name + "/" + id)
#         iter = self.collection.iter(ReadOptions(raw_mode=True))
#         iter.seek(key)

#         if not iter.key():
#             return {}

#         while iter.valid():
#             encoded_key = iter.key()
#             encoded_value = iter.value()

#             decoded_key = encoding.decode_str(encoded_key).split("/")
#             column = decoded_key[2]
#             document[column] = self._decode_value(encoded_value)

#             iter.next()

#         if not document:
#             return None

#         document["_id"] = id
#         return document

#     def _create_dir(self, dir_path: str, with_meta: bool = False):
#         if Path(dir_path).is_dir():
#             return False

#         # make directory
#         db_path = Path(dir_path)
#         db_path.mkdir(parents=True, exist_ok=True)

#         # make meta file
#         if with_meta:
#             with open(dir_path + "/meta.json", "w") as f:
#                 json.dump([], f, indent=4)

#         return True

#     def _delete_old_logs(self):
#         database_path = Path(self.db_path)
#         database_files = list(database_path.iterdir())

#         for filename in database_files:
#             if filename.name[:7] == "LOG.old":
#                 filename.unlink()

#     def _iterate_keys(self):
#         for key in self.collection.keys():
#             yield key

#         self._delete_old_logs()

#     def _add_index(self, index_specs: dict):
#         meta_file = self.db_path + "/full_text/meta.json"

#         with open(meta_file) as f:
#             index_data = json.load(f)

#         index_data.append(index_specs)

#         with open(meta_file, "w") as f:
#             json.dump(index_data, f, indent=4)

#     def _check_index_exists(self, index_name: str):
#         self._create_dir(self.db_path + "/full_text", with_meta=True)

#         if Path(self.db_path + "/full_text/" + index_name).is_dir():
#             return True

#         return False

#     def get_index(self, index_name: str, with_schema: bool = True):
#         fetched_schema = []
#         schema_builder = tantivy.SchemaBuilder()
#         schema_builder.add_text_field("_id", stored=True)

#         with open(self.db_path + "/full_text/meta.json") as f:
#             index_data = json.load(f)

#         for index in index_data:
#             if index["name"] == index_name:
#                 for field in index["schema"]:
#                     if field != "_id":
#                         schema_builder.add_text_field(field, stored=False)

#                 if with_schema:
#                     fetched_schema = index["schema"]

#                 index_path = index["path"]

#         # if not found_index: return None
#         schema = schema_builder.build()
#         index = tantivy.Index(schema, path=index_path)

#         if with_schema:
#             return index, fetched_schema

#         return index

#     def create(self, batch: bool = True):
#         if self._check_index_exists(self.name):
#             return self.get_index(self.name)

#         if not self.fields:
#             print(
#                 "Specify fields to index using the fields parameter i.e. create(name, fields=[])"
#             )
#             return None

#         index_path = self.db_path + "/full_text/" + self.name

#         index_specs = {"name": self.name, "schema": ["_id"], "path": index_path}

#         schema_builder = tantivy.SchemaBuilder()
#         schema_builder.add_text_field("_id", stored=True)

#         for field in self.fields:
#             schema_builder.add_text_field(field, stored=False)
#             index_specs["schema"].append(field)

#         schema = schema_builder.build()

#         self._create_dir(index_path)
#         self._add_index(index_specs)

#         index = tantivy.Index(schema, path=index_path)
#         writer = index.writer()

#         current_doc_id = ""
#         current_doc = {}

#         for key in self._iterate_keys():
#             decoded_key = encoding.decode_str(key).split("/")

#             doc_id = decoded_key[1]
#             key_column = decoded_key[2]

#             if doc_id != current_doc_id:
#                 if current_doc:
#                     # append doc to index
#                     current_doc["_id"] = [current_doc_id]
#                     writer.add_document(tantivy.Document(**current_doc))
#                     print(current_doc)
#                     if not batch:
#                         writer.commit()

#                 current_doc = {}
#                 current_doc_id = doc_id

#             if key_column in self.fields:
#                 key_value = self._get(key)

#                 if key_value:
#                     current_doc[key_column] = [key_value]

#         if batch:
#             writer.commit()

#         return index

#     def search(self, query: str, fields=None, limit: int = 1):
#         results = []

#         index, schema_fields = self.get_index(self.name, True)
#         if not fields:
#             fields = schema_fields

#         index.reload()
#         searcher = index.searcher()
#         parsed_query = index.parse_query(query, fields)
#         text_results = searcher.search(parsed_query, limit).hits

#         for result in text_results:
#             score, address = result
#             document_id = searcher.doc(address)["_id"][0]

#             results.append(self.get(document_id))

#         return results



import rockydb.encoding as encoding
from rocksdict import ReadOptions


class Index:
    def __init__(
        self, collection, collection_name: str, name: str, index_id: int, field: str
    ):
        self.collection = collection
        self.name = name
        self.field = field
        self.id = index_id
        self.collection_name = collection_name
        self.key_count = 0

    def _iter_default_db(self):
        key = encoding.encode_str(self.collection_name + "/0/0/")
        iter = self.collection.iter(ReadOptions(raw_mode=True))
        iter.seek(key)

        if not iter.key():
            return

        while iter.valid():
            encoded_key = iter.key()
            encoded_value = iter.value()
            decoded_key = encoding.decode_str(encoded_key).split("/")

            if decoded_key[0] != self.collection_name:
                break

            if decoded_key[4] == self.field:
                # returns (doc_id, bytes)
                yield (decoded_key[3], encoded_value)

            iter.next()

    def _iter_index_db(self, start: int = 0, limit: int = None):
        # returns doc_id
        if limit is None:
            limit = self.key_count - 1

        while 1:
            encoded_key = encoding.encode_str(f"{self.id}/{start}")
            decoded_key = encoding.decode_str(encoded_key).split("/")

            # make sure its not over the limit, to prevent key not found errors
            if int(decoded_key[1]) > limit:
                break

            encoded_value = self.collection[encoded_key]
            if not encoded_value:
                return

            doc_id = encoding.decode_str(encoded_value[1:])
            yield doc_id

            start += 1

    def create(self):
        # implement some sort algo that doesn't require all the data to be in memory
        # for now, brute force it; read 100 rows, insert in dict and sort, insert into dummy db
        i = 0
        block_id = 0

        # tmp/block_id/order_no/doc_id -> datatype_id/value
        block = {}
        for k, v in self._iter_default_db():
            tmp_key = f"{block_id}/{k}"
            block[tmp_key] = v
            i += 1

            if i == 100:
                block_sorted = dict(sorted(block.items(), key=lambda item: item[1]))
                block_rename = self._rename_block_keys(block_sorted)
                self._insert_tmp_kv(block_rename)
                block_id += 1
                i = 0
                block = {}

        # add remaining kv
        if block:
            block_sorted = dict(sorted(block.items(), key=lambda item: item[1]))
            block_renamed = self._rename_block_keys(block_sorted)
            self._insert_tmp_kv(block_renamed)
            block_id += 1

        # now merge sort all the blocks together
        self._merge_blocks(block_id)

        index_specs = {"name": self.name, "field": self.field, "id": self.id}
        return index_specs

    def _rename_block_keys(self, block: dict):
        result = {}
        i = 0

        for key, value in block.items():
            # current key: str block_id/doc_id, need to be tmp/block_id/order_no/doc_id
            split_key = key.split("/")
            new_key = f"tmp/{split_key[0]}/{i}/{split_key[1]}"

            result[new_key] = value
            i += 1

        return result

    def _insert_tmp_kv(self, kv_pairs: dict):
        # insert all blocks of sorted kv pairs back into db
        for k, v in kv_pairs.items():
            k = encoding.encode_str(k)
            self.collection[k] = v

    def _merge_blocks(self, block_count: int):
        # merge all blocks together, have pointers to start of blocks
        # take first key from each block, compare, insert smallest into new db
        iter = self.collection.iter(ReadOptions(raw_mode=True))

        base_block = 0

        # keep track of all pointer positions
        block_i_count = [0 for _ in range(block_count)]
        block_id_increment = 0

        while 1:
            # make sure base block is not complete
            if block_i_count[base_block] <= -1:
                for block_id in range(block_count):
                    if block_i_count[block_id] != -1:
                        base_block = block_id
                        block_id_increment = base_block
                        break

                if block_i_count[base_block] <= -1:
                    break

            base_key = encoding.encode_str(
                f"tmp/{base_block}/{block_i_count[base_block]}/"
            )
            iter.seek(base_key)
            k = iter.key()
            k_value = iter.value()

            for block_id in range(block_count):
                # for each block, start at first key and iterate through all other 99 keys
                # check if block is complete, if so, skip it
                if block_i_count[block_id] <= -1:
                    continue

                block_key = encoding.encode_str(
                    f"tmp/{block_id}/{block_i_count[block_id]}/"
                )
                iter.seek(block_key)
                block_k_value = iter.value()
                block_key = iter.key()

                # iter.seek() will go to the next key if the key doesn't exist, however we want to skip the block if the key doesn't exist
                should_be_key = f"tmp/{block_id}/{block_i_count[block_id]}"
                if (
                    encoding.decode_str(block_key)[: len(should_be_key)]
                    != should_be_key
                ):
                    block_i_count[block_id] = -2
                    continue

                # this should be fixed in earlier stages, set None to 0
                if block_k_value is None:
                    continue

                # if block value is shorter than the base value, then insert that instead
                if block_k_value < k_value:
                    k = block_key
                    k_value = block_k_value
                    block_id_increment = block_id

            # increment pointer for the block we just go the value from
            block_i_count[block_id_increment] += 1

            if block_i_count[block_id_increment] != -1:
                # found the smallest key between all pointers, insert into new db,
                # index_id/order_no -> str/doc_id
                new_key = encoding.encode_str(f"{self.id}/{self.key_count}")
                decoded_doc_id = encoding.decode_str(k).split("/")[3]
                encoded_doc_id = encoding.encode_str(decoded_doc_id)
                encoded_data_type = encoded_data_type = encoding.encode_int(
                    1
                )  # encode id for str is 1

                self.collection[new_key] = encoded_data_type + encoded_doc_id
                self.key_count += 1

            # if block is complete, set to -1
            if block_i_count[block_id_increment] == 100:
                block_i_count[block_id_increment] = -1

                # if block is complete, use another block as the new base key
                found_new_block = 0
                for i in range(block_count):
                    if block_i_count[i] != -1:
                        base_block = block_id_increment
                        found_new_block = 1
                        break

                if not found_new_block:
                    # we have iterated through all blocks, no new blocks have been found all keys have been inserted in order
                    break

            else:
                # otherwise, set the new base key to be from the block that we inserted from
                base_block = block_id_increment

    def _parse_query(self, query: dict):
        lt = 0
        lte = 0
        gt = 0
        gte = 0
        eq = 0
        # need to iterate over keys and values once only
        for spec, v in query.items():
            q_loc = spec.find("?")

            if q_loc != -1:
                query_type = spec[q_loc + 1 :]

                # can only be of one type below
                if query_type == "lte":
                    lte = encoding.encode_int(v)
                    continue

                if query_type == "gte":
                    gte = encoding.encode_int(v)
                    continue

                if query_type == "lt":
                    lt = encoding.encode_int(v)
                    continue

                if query_type == "gt":
                    gt = encoding.encode_int(v)
                    continue

            else:
                # add to equals dict
                eq[spec] = v

        return (lt, lte, gt, gte, eq)

    def find(self, query: dict, limit: int = 1):
        results = []
        if not query or not limit:
            return results

        lt, lte, gt, gte, eq = self._parse_query(query)

        # if query is less than, e.g. { "age?lte":  30 }
        if lte:
            index_id = self.less_than_equals(lte)
            for doc_id in self._iter_index_db(start=0, limit=index_id):
                results.append(doc_id)

            return results

        if gte:
            index_id = self.greater_than_equals(gte)
            for doc_id in self._iter_index_db(start=index_id, limit=None):
                results.append(doc_id)

            return results

        if lt:
            index_id = self.less_than(lt)
            for doc_id in self._iter_index_db(start=0, limit=index_id):
                results.append(doc_id)

            return results

        if gt:
            index_id = self.greater_than(gt)
            for doc_id in self._iter_index_db(start=index_id, limit=None):
                results.append(doc_id)

            return results

        return results

    def _get_value_from_index_key(self, key: str) -> bytes:
        index_key = encoding.encode_str(key)
        index_value = self.collection[index_key]
        doc_id = encoding.decode_str(index_value[1:])

        db_key = encoding.encode_str(
            f"{self.collection_name}/0/0/{doc_id}/{self.field}"
        )
        db_value = self.collection[db_key]

        return db_value

    def less_than_equals(self, target: bytes) -> int:
        # for lte type search, also know as last occurance
        min = 0
        max = self.key_count

        while min <= max:
            mid = (min + max) // 2
            mid_value = self._get_value_from_index_key(f"{self.id}/{mid}")[
                1:
            ]  # bytes of value
            mid_value_plus_one = self._get_value_from_index_key(f"{self.id}/{mid + 1}")[
                1:
            ]  # bytes of value

            if ((mid == self.key_count) or (mid_value_plus_one > target)) and (
                mid_value == target
            ):
                return mid

            elif target < mid_value:
                max = mid - 1

            else:
                min = mid + 1

        if min < max:
            return min
        return max

    def greater_than_equals(self, target: bytes) -> int:
        # for gte type search, also known as first occurance
        min = 0
        max = self.key_count

        while min < max:
            mid = (min + max) // 2
            mid_value = self._get_value_from_index_key(f"{self.id}/{mid}")[1:]
            mid_value_minus_one = self._get_value_from_index_key(
                f"{self.id}/{mid - 1}"
            )[1:]

            if ((mid == self.key_count) or (mid_value_minus_one < target)) and (
                mid_value == target
            ):
                return mid

            elif target > mid_value:
                min = mid + 1

            else:
                max = mid

        return min

    def less_than(self, target: bytes) -> int:
        # for lt type search
        min = 0
        max = self.key_count

        while min < max:
            mid = (min + max) // 2
            mid_value = self._get_value_from_index_key(f"{self.id}/{mid}")[1:]
            mid_value_plus_one = self._get_value_from_index_key(f"{self.id}/{mid + 1}")[
                1:
            ]

            if ((mid == self.key_count) or (mid_value_plus_one >= target)) and (
                mid_value < target
            ):
                return mid

            elif mid_value >= target:
                max = mid - 1

            else:
                min = mid + 1

        return min

    def greater_than(self, target: bytes) -> int:
        # for gt type search
        min = 0
        max = self.key_count

        while min < max:
            mid = (min + max) // 2
            mid_value = self._get_value_from_index_key(f"{self.id}/{mid}")[1:]
            mid_value_minus_one = self._get_value_from_index_key(
                f"{self.id}/{mid - 1}"
            )[1:]

            if (mid_value_minus_one <= target) and (mid_value > target):
                return mid

            elif mid_value <= target:
                min = mid + 1

            else:
                max = mid

        return min

    def get(self, id: str) -> dict:
        document = {}

        for encoded_key in self._id_rows(id):
            decoded_key = encoding.decode_str(encoded_key).split("/")

            column = decoded_key[4]
            document[column] = self._get(encoded_key)

        if not document:
            return None

        document["_id"] = id
        return document

    def _id_rows(self, id: str):
        key = encoding.encode_str(self.collection_name + "/0/0/" + id)
        iter = self.collection.iter(ReadOptions(raw_mode=True))
        iter.seek(key)

        if not iter.key():
            return {}

        while iter.valid():
            encoded_key = iter.key()
            decoded_key = encoding.decode_str(encoded_key).split("/")
            if decoded_key[3] != id:
                break

            yield encoded_key
            iter.next()

    def _decode_value(self, value: bytes):
        if not value:
            return None

        decoded_value = encoding.decode_this(str, value[1:])
        return decoded_value

    def _get(self, key: bytes):
        value = self.collection[key]
        return self._decode_value(value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rockydb']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.85.1,<0.86.0',
 'pytest>=7.2.0,<8.0.0',
 'rocksdict==0.3.2',
 'uvicorn>=0.19.0,<0.20.0']

setup_kwargs = {
    'name': 'rockydb',
    'version': '0.2.10',
    'description': 'A NoSQL database.',
    'long_description': '# RockyDB \n[![License](https://img.shields.io/badge/License-Apache_2.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n![CI](https://github.com/aaldulimi/rockydb/actions/workflows/integrate.yml/badge.svg)\n[![codecov](https://codecov.io/github/aaldulimi/RockyDB/branch/master/graph/badge.svg?token=6MZLCKX5IJ)](https://codecov.io/github/aaldulimi/RockyDB)\n\nSimple NoSQL database written in Python. It relies on rocksdb as its storage engine. This is more of a Proof-of-concept than a production-ready database. \n\n## Installation \n```\npip install rockydb\n```\n\n## Contents\n- [RockyDB](#rockydb)\n- [Installation](#installation)\n- [Contents](#contents)\n- [Features](#features)\n- [Documentation](#documentation)\n    - [Create collection](#create-collection)\n    - [Insert doucment](#insert-document)\n    - [Get document](#get-document)\n    - [Delete document](#delete-document)\n    - [Query](#query)\n    \n\n\n## Features\nCurrently under active development, however here is the feature list so far:\n\n- **Create collections**\n- **Insert, get and delete documents**\n- **REST API**\n- **Query language**\n- **Indexes**\n- **Full-text Search [IN-DEVELOPMENT]**\n\n## Performance\nDataset: [NBA Players Dataset](https://www.kaggle.com/datasets/drgilermo/nba-players-stats).\nComputer: MacBook Pro (13-inch, 2019).\nRockyDB is still in its early days, these results will likely get better in the future. \n| Database      | Insert | Get | Query | Delete \n| -----------| -----------:| -----------:| -----------:| -----------:| \n| RockyDB      | **0.00074**       | **0.00038** | 0.00014 | **0.00023**\n| MongoDB   | 0.04436        | 0.04518 | **0.00004**  | 0.04264\n\n## Documentation\nFull [Documentation](https://rockydb.readthedocs.io/en/latest/). Below are the basics:\n### Create collection \n```python\nfrom rockydb import RockyDB\n\ndb = RockyDB(path="database/")\nnews = db.collection("news")\n```\n\n### Insert document\nSupported data types: `str`, `int`, `float`, `bool` and `list`. Will support more later. \n```python\ndoc_id = news.insert({\n  "title": "Can store strings",\n  "year": 2022,\n  "people": ["lists", "are", "fine", "too"],\n  "pi": 3.14,\n  "real": True\n})\n```\nThe `insert` method will return a unique document `_id`. `_id` will be created if document does not contain it.  \n\n### Get document\n```python\nnews.get(doc_id)\n```\n### Delete document\n```python\nnews.delete(doc_id)\n```\n### Query\n```python\nnews.find({"pi?lt": 3.14, "real": True}, limit=10)\n``` \nThe `limit` arg is optional, default is 10. Supports exact, lte, lt, gt and gte queries.\n',
    'author': 'Ahmed',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aaldulimi/RockyDB',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from pathlib import Path
import json
import string
from rocksdict import (
    Rdict,
    Options,
    ReadOptions,
    WriteBatch,
    CompactOptions,
)
import random
from rockydb.index import Index
import rockydb.encoding as encoding
import os


class Collection:
    def __init__(self, db_path: str, name: str):
        self.db_path = db_path
        self.name = name
        self.path = self.db_path + name

        self.opt = Options(raw_mode=True)
        self.opt.increase_parallelism(os.cpu_count())
        self.opt.set_allow_mmap_reads(True)
        self.opt.set_write_buffer_size(0x10000000)

        self._create_dir(self.path, with_meta=True)
        self.collection = Rdict(path=self.path, options=self.opt)

        self.encoding_types = {
            str: 1,
            int: 2,
            float: 3,
            bool: 4,
            list: 5,
            1: str,
            2: int,
            3: float,
            4: bool,
            5: list,
        }

    def _create_dir(self, dir_path: str, with_meta: bool = False):
        if Path(dir_path).is_dir():
            return False

        # make directory
        db_path = Path(dir_path)
        db_path.mkdir(parents=True, exist_ok=True)

        # make meta file
        if with_meta:
            with open(dir_path + "/meta.json", "w") as f:
                json.dump([], f, indent=4)

        return True

    def _delete_old_logs(self):
        database_path = Path(self.path)
        database_files = list(database_path.iterdir())

        for filename in database_files:
            if filename.name[:7] == "LOG.old":
                filename.unlink()

    def _generate_id(self):
        characters = string.ascii_letters + string.digits
        doc_id = "".join(random.choice(characters) for _ in range(8))

        return doc_id

    def insert_object(self, document: object) -> str:
        document_dict = document.__dict__.copy()
        doc_id = self.insert(document_dict)

        return doc_id

    def insert(self, document: dict, wb: WriteBatch = None) -> str:
        # encoding method
        # collection_id/index_id/order_no/doc_id/col_id -> datatype_id/value
        # order_no is for sorting, will be 0 for default index

        if "_id" not in document:
            document["_id"] = self._generate_id()

        doc_id = document["_id"]

        for key, value in document.items():
            if key != "_id":
                key_string = f"{self.name}/0/0/{doc_id}/{key}"
                encoded_data = encoding.encode_this(value)
                encoded_data_type = encoding.encode_int(
                    self.encoding_types[type(value)]
                )  # byte of length 1

                encoded_key = encoding.encode_str(key_string)
                encoded_value = encoded_data_type + encoded_data

                # insert in db
                if wb is not None:
                    wb[encoded_key] = encoded_value
                else:
                    self.collection[encoded_key] = encoded_value
                    # print(f"{encoded_key} -> {encoding.decode_int(encoded_value)}")

        self._delete_old_logs()
        return doc_id

    def insert_batch(self, document_list: list):
        wb = WriteBatch(raw_mode=True)

        for document in document_list:
            self.insert(document, wb)

        self.collection.write(wb)

    def insert_object_batch(self, object_list: list):
        for object in object_list:
            self.insert_object(object)

    def _decode_value(self, value: bytes):
        if not value:
            return None

        decoded_data_type = self.encoding_types[value[0]]
        decoded_value = encoding.decode_this(decoded_data_type, value[1:])

        return decoded_value

    def _get(self, key: bytes):
        value = self.collection[key]
        return self._decode_value(value)

    def _iterate_keys(self):
        for key in self.collection.keys():
            yield key

        self._delete_old_logs()

    # def create_index(self, name: str, fields: list, batch: bool = False):
    #     index = Index(
    #         self.path,
    #         self.collection,
    #         self.name,
    #         name,
    #         fields,
    #         encoding_types=self.encoding_types,
    #     )
    #     index.create(batch)

    #     return index

    # def get_index(self, name: str):
    #     index = Index(
    #         self.path,
    #         self.collection,
    #         self.name,
    #         name,
    #         encoding_types=self.encoding_types,
    #     )
    #     index.get_index(name)

    #     return index

    def get_id_contains(self, field: str, value, max_count: int = None):
        all_ids = []

        for key in self._iterate_keys():
            decoded_key = encoding.decode_str(key).split("/")
            key_column = decoded_key[4]
            key_value = self._get(key)

            if (field == key_column) and key_value:
                if value in self._get(key):
                    row_id = decoded_key[3]
                    all_ids.append(row_id)

            if max_count:
                if len(all_ids) == max_count:
                    return all_ids

        return all_ids

    def _contains(self, field: str, value, max_count: int = None):
        results = []
        doc_ids = self.get_id_contains(field, value, max_count)

        if doc_ids:
            for doc_id in doc_ids:
                doc_dict = {}
                doc_dict["_id"] = doc_id

                for key in self._iterate_keys():
                    decoded_key = encoding.decode_str(key).split("/")
                    search_doc_id = decoded_key[3]

                    if search_doc_id == doc_id:

                        column_name = decoded_key[4]
                        doc_dict[column_name] = self._get(key)

                results.append(doc_dict)

        return results

    def find(self, query: dict, limit: int = 1):
        results = []
        if not query or not limit:
            return results

        lt = {}
        lte = {}
        gt = {}
        gte = {}
        eq = {}
        # need to iterate over keys and values once only
        for spec, v in query.items():
            q_loc = spec.find("?")

            if q_loc != -1:
                query_type = spec[q_loc + 1 :]

                # can only be of one type below
                if query_type == "lte":
                    lte[spec[:q_loc]] = v
                    continue

                if query_type == "gte":
                    gte[spec[:q_loc]] = v
                    continue

                if query_type == "lt":
                    lt[spec[:q_loc]] = v
                    continue

                if query_type == "gt":
                    gt[spec[:q_loc]] = v
                    continue

                print(f"Unkown query type {query_type}")
                return results

            else:
                # add to equals dict
                eq[spec] = v

        # iterate through all keys to find doc ids that match
        count = 0
        read_opt = ReadOptions(raw_mode=True)
        read_opt.fill_cache(False)
        read_opt.set_readahead_size(8_388_608)
        read_opt.set_tailing(True)
        read_opt.set_pin_data(True)

        for k, v in self.collection.items(read_opt=read_opt):
            decoded_key = encoding.decode_str(k).split("/")
            column = decoded_key[4]

            # check for equal
            if column in eq:
                if query[column] == self._decode_value(v):
                    results.append(self.get(decoded_key[3]))
                    count += 1

            if column in lte:
                if self._decode_value(v) <= lte[column]:
                    results.append(self.get(decoded_key[3]))
                    count += 1

            if column in gte:
                if self._decode_value(v) >= gte[column]:
                    results.append(self.get(decoded_key[3]))
                    count += 1

            if column in lt:
                if self._decode_value(v) < lt[column]:
                    results.append(self.get(decoded_key[3]))
                    count += 1

            if column in gt:
                if self._decode_value(v) > gt[column]:
                    results.append(self.get(decoded_key[3]))
                    count += 1

            if count == limit:
                break

        return results

    def delete(self, id: str) -> bool:
        found_doc = False
        writebatch = WriteBatch(raw_mode=True)

        for encoded_key in self._id_rows(id):
            writebatch.delete(encoded_key)
            found_doc = True

        self.collection.write(writebatch)

        return found_doc

    def _id_rows(self, id: str):
        key = encoding.encode_str(self.name + "/0/0/" + id)
        iter = self.collection.iter(ReadOptions(raw_mode=True))
        iter.seek(key)

        if not iter.key():
            return {}

        while iter.valid():
            encoded_key = iter.key()
            decoded_key = encoding.decode_str(encoded_key).split("/")
            if decoded_key[3] != id:
                break

            yield encoded_key
            iter.next()

    def delete_batch(self, id_list: list):
        for id in id_list:
            self.delete(id)

    def _delete_tmp_blocks(self):
        iter = self.collection.iter(ReadOptions(raw_mode=True))
        iter.seek(b"tmp/")

        while iter.valid():
            k = iter.key()
            if k[:3] != b"tmp":
                break

            self.delete(encoding.decode_str(k))
            iter.next()

    def get(self, id: str) -> dict:
        document = {}

        for encoded_key in self._id_rows(id):
            decoded_key = encoding.decode_str(encoded_key).split("/")

            column = decoded_key[4]
            document[column] = self._get(encoded_key)

        if not document:
            return None

        document["_id"] = id
        return document

    def get_batch(self, id_list: list):
        results = []

        for id in id_list:
            document = self.get(id)
            results.append(document)

        return results

    def create_index(self, name: str, field: str):
        index_id = 0
        # check if index already exists
        meta_file = self.path + "/meta.json"
        with open(meta_file, "r") as f:
            index_data = json.load(f)

        for index in index_data:
            if index["name"] == name:
                print("Index already exists, please choose a different index name")
                return None

            if index["id"] >= index_id:
                index_id = index["id"]

        index = Index(self.collection, self.name, name, index_id + 1, field)
        index_spec = index.create()

        # now delete all tmp blocks that were saved in db
        self._delete_tmp_blocks()

        # update meta file once index fully created
        index_data.append(index_spec)
        with open(self.path + "/meta.json", "w") as f:
            json.dump(index_data, f, indent=4)

        return index

    def get_index(self, name: str):
        pass

    def delete_index(self, name: str):
        pass

    def destroy(self):
        Rdict.destroy(self.path)

    def flush(self, wait: bool = False):
        self.collection.flush(wait)

    def compact_range(self, start: bytes = None, end: bytes = None):
        if not start or not end:
            iter = self.collection.iter(ReadOptions(raw_mode=True))
            iter.seek_to_first()
            start = iter.key()
            iter.seek_to_last()
            end = iter.key()

        self.collection.compact_range(start, end, CompactOptions())
